'***************************************************************************************
' This program demonstrates the use of an LMM interpreter within the Spin interpreter.
' It calls the lmm.start routine, which patches the LMM interpreter into Spin interpreter.
' The sqrt, strsize and the strcomp routines are replaced by LMM PASM routines that run
' at about the same speed as the original routines.
'
' All of the code in this program executes in a single cog.  The speeds of the sqrt,
' strsize and strcomp routines are measured before and after starting the LMM PASM
' interpreter.  The speeds of a simple summing loop is compared between Spin and two
' LMM PASM routines, where on routine uses the FCACHE feature.
'
' This program includes the HalfDuplexSerial object, which also uses LMM PASM to run
' in the same cog.
'
' Dave Hein, July 1, 2010
'***************************************************************************************
CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

OBJ
  lmm : "lmm_interp"
  hds : "HalfDuplexSerial"

CON
  ' LMM PASM Constants
  x           = lmm#x
  y           = lmm#y
  fcache      = lmm#fcache
  fcall       = lmm#fcall
  lmm_ret     = lmm#lmm_ret
  cache_addr  = lmm#cache_addr
  fjmp        = lmm#fjmp
  lmm_loop    = lmm#lmm_loop
  lmm_pc      = lmm#lmm_pc
  fretx       = lmm#fretx

PUB main | regnum, time1[3], time2[3], instr[20]

  waitcnt(clkfreq+cnt)

  ' Run time tests before adding LMM PASM interpreter
  TimeTest(@time1)

  ' Start the LMM interpreter
  lmm.start

  ' Run time tests after adding LMM PASM interpreter
  TimeTest(@time2)

  hds.tx(13)
  hds.dbprintf2(string("sqrt cycles = %d, %d\n"), time1[0], time2[0])
  hds.dbprintf2(string("strsize cycles = %d, %d\n"), time1[1], time2[1])
  hds.dbprintf2(string("strcomp cycles = %d, %d\n"), time1[2], time2[2])

  repeat
    hds.dbprintf0(string("\nEnter a number: "))
    hds.getstr(@instr, 80)
    regnum := hds.strtodec(@instr)
    if regnum > 0
      SubroutineTest(regnum)

' This routine calls the three routines that implement a simple summing loop.  The routines are
' written in Spin, LMM PASM and LMM PASM using the FCACHE psuedo-op
PUB SubroutineTest(num) | sum, cycles
  cycles := cnt
  sum := SpinSum(num)
  cycles := cnt - cycles
  hds.dbprintf2(string("The SpinSum routine returned %d at %d cycles per loop\n"), sum, cycles/num)

  cycles := cnt
  sum := lmm.run(@LmmSum1, num)
  cycles := cnt - cycles
  hds.dbprintf2(string("The LmmSum1 routine returned %d at %d cycles per loop\n"), sum, cycles/num)

  cycles := cnt
  sum := lmm.run(@LmmSum2, num)
  cycles := cnt - cycles
  hds.dbprintf2(string("The LmmSum2 routine returned %d at %d cycles per loop\n"), sum, cycles/num)

' Sum the integers from 1 to num
PUB SpinSum(num) | i
  repeat i from 1 to num
    result += i

DAT
' This LMM PASM routine computes the sum of the numbers from 1 to N.  It uses FJMP to jump back to
' the beginning of the loop.  FCALL is used to demonstrate how calls are made in LMM PASM.
              org       0
LmmSum1       jmp       #FCALL            ' Call the routine at the following address
              long      @@@LmmSum1a
              jmp       #FRETX            ' Push x to the stack and return to Spin

LmmSum1a      mov       y, #0
LmmSum1b      add       y, x
              djnz      x, #FJMP          ' Jump to the following address
              long      @@@LmmSum1b
              mov       x, y
              mov       lmm_pc, lmm_ret   ' Return to the LMM PASM caller

' This LMM PASM routine uses FCACHE to execute a loop that sums the integers from 1 to N
LmmSum2       jmp       #FCALL            ' Call the routine at the following address
              long      @@@LmmSum2a
              jmp       #FRETX            ' Push x to the stack and return to Spin

LmmSum2a      mov       y, #0
              jmp       #FCACHE           ' Load the following PASM code and jump to it
              org       cache_addr        ' Reset Cog addres to cache address
LmmSum2b      add       y, x
              djnz      x, #LmmSum2b
              jmp       #lmm_loop         ' Jump back to the LMM PASM interpreter
              long      0                 ' Terminate FCACHE instruction with a zero
              mov       x, y
              mov       lmm_pc, lmm_ret   ' Return to the LMM PASM caller

' This routine runs timing tests on sqrt, strsize and strcomp
PUB TimeTest(pTimes) | cycles, ptr

  ptr :=  string("1234567890123456789012345")

  cycles := cnt
  result := ^^ 169
  cycles := cnt - cycles
  long[pTimes][0] := cycles

  cycles := cnt
  result := strsize(ptr)
  cycles := cnt - cycles
  long[pTimes][1] := cycles

  cycles := cnt
  result := strcomp(ptr, ptr)
  cycles := cnt - cycles
  long[pTimes][2] := cycles

